import json
import math
from pathlib import Path

import torch

from .kitsune import engine
from .kitsune.data import FileFormat, build_input_data_pipe
from .kitsune.models import Kitsune

device = torch.device("cuda" if torch.cuda.is_available() else "cpu")


def compute_anomaly_scores(
    path: Path,
    scores_dir: Path,
    batch_size: int = 32,
    n_scores_partition: int = 4,
    file_format: FileFormat = "csv",
    model_checkpoint: Path = Path("models/kitsune.pt")
) -> None:

    scores_dir.mkdir(exist_ok=True, parents=True)

    print("🦊 Loading model...")
    model = Kitsune.from_pretrained(model_checkpoint)
    model.to(device)

    print("🦊 Building input data pipeline...")
    dp = build_input_data_pipe(root=str(path),
                               batch_size=batch_size,
                               file_format=file_format,
                               shuffle=False)

    print("🦊 Running inference...")
    scores = engine.predict(model, dp, device=device)

    print("🦊 Serializing scores...")
    scores_chunks = scores.split(math.ceil(len(scores) / n_scores_partition))
    for i, chunk in enumerate(scores_chunks):
        fname = scores_dir / f"scores_{i}.json"
        print(f">> Serializing scores chunk {i} ({fname})...")

        with fname.open("w") as score_f:
            lines = [json.dumps({"score": s}) for s in chunk.cpu().tolist()]
            score_f.write("\n".join(lines))



from setuptools import setup , find_packages 

classifiers = []

setup(
    name='tarda',
    version='0.0.7',
    description='A library for collection of IDS & robust-IDS and tools for evaluating them',
    long_description=open("README.md").read() + '\n\n' + open("CHANGELOG.txt").read(),
    url='https://github.com/spg-iitd/tardigrade',
    author='Swain Subrat Kumar',
    author_email='mailofswainsubrat@gmail.com',
    license='MIT',
    include_package_data=True,
    # classifiers=classifiers,
    keywords='ids adversarial network nids',
    packages= find_packages(),
    install_requires=['scapy', 'numpy', 'pandas', 'matplotlib', 'scikit-learn', 'scipy', 'tqdm', 'torch', 'torchvision', 'torchsummary', 'torchattacks'],
    zip_safe=False
)



# Copyright (c) 2017 Yusuke Sugomori
#
# MIT License
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
# LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
# OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
# WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

# Portions of this code have been adapted from Yusuke Sugomori's code on GitHub: https://github.com/yusugomori/DeepLearning

import sys
import numpy
from .utils import *
import json

def squeeze_features(fv, precision):
    """rounds features to siginificant figures

    Args:
        fv (array): feature vector.
        precision (int): number of precisions to use.

    Returns:
        array: rounded array of floats.

    """

    return numpy.around(fv, decimals=precision)

def quantize(x, k):
    n = 2**k - 1
    return numpy.round(numpy.multiply(n, x))/n


def quantize_weights(w, k):
    x = numpy.tanh(w)
    q = x / numpy.max(numpy.abs(x)) * 0.5 + 0.5
    return 2 * quantize(q, k) - 1


class dA_params:
    def __init__(self, n_visible=5, n_hidden=3, lr=0.001, corruption_level=0.0, gracePeriod=10000, hiddenRatio=None, normalize=True, input_precision=None, quantize=None):
        self.n_visible = n_visible  # num of units in visible (input) layer
        self.n_hidden = n_hidden  # num of units in hidden layer
        self.lr = lr
        self.corruption_level = corruption_level
        self.gracePeriod = gracePeriod
        self.hiddenRatio = hiddenRatio
        self.normalize = normalize
        self.quantize=quantize
        self.input_precision=input_precision
        if quantize:
            self.q_wbit,self.q_abit=quantize


class dA:
    def __init__(self, params):
        self.params = params

        if self.params.hiddenRatio is not None:
            self.params.n_hidden = int(numpy.ceil(
                self.params.n_visible * self.params.hiddenRatio))

        # for 0-1 normlaization
        self.norm_max = numpy.ones((self.params.n_visible,)) * -numpy.Inf
        self.norm_min = numpy.ones((self.params.n_visible,)) * numpy.Inf
        self.n = 0

        self.rng = numpy.random.RandomState(1234)

        a = 1. / self.params.n_visible
        self.W = numpy.array(self.rng.uniform(  # initialize W uniformly
            low=-a,
            high=a,
            size=(self.params.n_visible, self.params.n_hidden)))

        #quantize weights
        if self.params.quantize:
            self.W=quantize_weights(self.W, self.params.q_wbit)

        self.hbias = numpy.zeros(self.params.n_hidden)  # initialize h bias 0
        self.vbias = numpy.zeros(self.params.n_visible)  # initialize v bias 0
        # self.W_prime = self.W.T

    def get_corrupted_input(self, input, corruption_level):
        assert corruption_level < 1

        return self.rng.binomial(size=input.shape,
                                 n=1,
                                 p=1 - corruption_level) * input

    # Encode
    def get_hidden_values(self, input):
        return sigmoid(numpy.dot(input, self.W) + self.hbias)

    # Decode
    def get_reconstructed_input(self, hidden):
        return sigmoid(numpy.dot(hidden, self.W.T) + self.vbias)

    def train(self, x):
        self.n = self.n + 1

        if self.params.normalize:
            # update norms
            self.norm_max[x > self.norm_max] = x[x > self.norm_max]
            self.norm_min[x < self.norm_min] = x[x < self.norm_min]

            # 0-1 normalize
            x = (x - self.norm_min) / (self.norm_max -
                                       self.norm_min + 0.0000000000000001)

        if self.params.input_precision:
            x=squeeze_features(x,self.params.input_precision)

        if self.params.corruption_level > 0.0:
            tilde_x = self.get_corrupted_input(x, self.params.corruption_level)
        else:
            tilde_x = x

        y = self.get_hidden_values(tilde_x)
        if self.params.quantize:
            y=quantize(y, self.params.q_abit)

        z = self.get_reconstructed_input(y)

        L_h2 = x - z
        L_h1 = numpy.dot(L_h2, self.W) * y * (1 - y)

        L_vbias = L_h2
        L_hbias = L_h1
        L_W = numpy.outer(tilde_x.T, L_h1) + numpy.outer(L_h2.T, y)

        self.W += self.params.lr * L_W
        self.hbias += self.params.lr * L_hbias
        self.vbias += self.params.lr * L_vbias

        if self.params.quantize:
            self.W=quantize_weights(self.W, self.params.q_wbit)
            self.hbias=quantize_weights(self.hbias, self.params.q_wbit)
            self.vbias=quantize_weights(self.vbias, self.params.q_wbit)
        # the RMSE reconstruction error during training
        return numpy.sqrt(numpy.mean(L_h2**2))

    def reconstruct(self, x):
        y = self.get_hidden_values(x)

        try:
            if self.params.quantize:
                y=quantize(y, self.params.q_abit)
        except AttributeError as e:
            pass
            
        z = self.get_reconstructed_input(y)
        return z

    def get_params(self):
        params={
        "W":self.W,
        "hbias":self.hbias,
        "vbias":self.vbias
        }
        return params

    def set_params(self, new_param):
        self.W=new_param["W"]
        self.hbias=new_param["hbias"]
        self.vbias=new_param["vbias"]

    def execute(self, x):  # returns MSE of the reconstruction of x
        if self.n < self.params.gracePeriod:
            return 0.0
        else:
            # 0-1 normalize
            try:
                if self.params.normalize:
                    x = (x - self.norm_min) / (self.norm_max -
                                               self.norm_min + 0.0000000000000001)

                if self.params.input_precision:
                    x=squeeze_features(x,self.params.input_precision)

            except AttributeError as e:
                pass

            z = self.reconstruct(x)
            rmse = numpy.sqrt(((x - z) ** 2).mean())  # MSE
            return rmse

    def inGrace(self):
        return self.n < self.params.gracePeriod



import numpy as np
## Prep AfterImage cython package
import os
import subprocess
# import pyximport
# pyximport.install()
from .AfterImage import *
#import AfterImage_NDSS as af

#
# MIT License
#
# Copyright (c) 2018 Yisroel mirsky
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.


class netStat:
    #Datastructure for efficent network stat queries
    # HostLimit: no more that this many Host identifiers will be tracked
    # HostSimplexLimit: no more that this many outgoing channels from each host will be tracked (purged periodically)
    # Lambdas: a list of 'window sizes' (decay factors) to track for each stream. nan resolved to default [5,3,1,.1,.01]
    def __init__(self, Lambdas = np.nan, HostLimit=255,HostSimplexLimit=1000):
        #Lambdas
        if np.isnan(Lambdas):
            self.Lambdas = [5,3,1,.1,.01]
        else:
            self.Lambdas = Lambdas

        #HT Limits
        self.HostLimit = HostLimit
        self.SessionLimit = HostSimplexLimit*self.HostLimit*self.HostLimit #*2 since each dual creates 2 entries in memory
        self.MAC_HostLimit = self.HostLimit*10

        #HTs
        self.HT_jit = incStatDB(limit=self.HostLimit*self.HostLimit)#H-H Jitter Stats
        self.HT_MI = incStatDB(limit=self.MAC_HostLimit)#MAC-IP relationships
        self.HT_H = incStatDB(limit=self.HostLimit) #Source Host BW Stats
        self.HT_Hp = incStatDB(limit=self.SessionLimit)#Source Host BW Stats


    def findDirection(self,IPtype,srcIP,dstIP,eth_src,eth_dst): #cpp: this is all given to you in the direction string of the instance (NO NEED FOR THIS FUNCTION)
        if IPtype==0: #is IPv4
            lstP = srcIP.rfind('.')
            src_subnet = srcIP[0:lstP:]
            lstP = dstIP.rfind('.')
            dst_subnet = dstIP[0:lstP:]
        elif IPtype==1: #is IPv6
            src_subnet = srcIP[0:round(len(srcIP)/2):]
            dst_subnet = dstIP[0:round(len(dstIP)/2):]
        else: #no Network layer, use MACs
            src_subnet = eth_src
            dst_subnet = eth_dst

        return src_subnet, dst_subnet

    def updateGetStats(self, IPtype, srcMAC,dstMAC, srcIP, srcProtocol, dstIP, dstProtocol, datagramSize, timestamp):
        # Host BW: Stats on the srcIP's general Sender Statistics
        # Hstat = np.zeros((3*len(self.Lambdas,)))
        # for i in range(len(self.Lambdas)):
        #     Hstat[(i*3):((i+1)*3)] = self.HT_H.update_get_1D_Stats(srcIP, timestamp, datagramSize, self.Lambdas[i])

        #MAC.IP: Stats on src MAC-IP relationships
        MIstat =  np.zeros((3*len(self.Lambdas,)))
        for i in range(len(self.Lambdas)):
            MIstat[(i*3):((i+1)*3)] = self.HT_MI.update_get_1D_Stats(srcMAC+srcIP, timestamp, datagramSize, self.Lambdas[i])

        # Host-Host BW: Stats on the dual traffic behavior between srcIP and dstIP
        HHstat =  np.zeros((7*len(self.Lambdas,)))
        for i in range(len(self.Lambdas)):
            HHstat[(i*7):((i+1)*7)] = self.HT_H.update_get_1D2D_Stats(srcIP, dstIP,timestamp,datagramSize,self.Lambdas[i])

        # Host-Host Jitter:
        HHstat_jit =  np.zeros((3*len(self.Lambdas,)))
        for i in range(len(self.Lambdas)):
            HHstat_jit[(i*3):((i+1)*3)] = self.HT_jit.update_get_1D_Stats(srcIP+dstIP, timestamp, 0, self.Lambdas[i],isTypeDiff=True)

        # Host-Host BW: Stats on the dual traffic behavior between srcIP and dstIP
        HpHpstat =  np.zeros((7*len(self.Lambdas,)))
        if srcProtocol == 'arp':
            for i in range(len(self.Lambdas)):
                HpHpstat[(i*7):((i+1)*7)] = self.HT_Hp.update_get_1D2D_Stats(srcMAC, dstMAC, timestamp, datagramSize, self.Lambdas[i])
        else:  # some other protocol (e.g. TCP/UDP)
            for i in range(len(self.Lambdas)):
                HpHpstat[(i*7):((i+1)*7)] = self.HT_Hp.update_get_1D2D_Stats(srcIP + srcProtocol, dstIP + dstProtocol, timestamp, datagramSize, self.Lambdas[i])

        return np.concatenate((MIstat, HHstat, HHstat_jit, HpHpstat))  # concatenation of stats into one stat vector

    def getNetStatHeaders(self):
        MIstat_headers = []
        Hstat_headers = []
        HHstat_headers = []
        HHjitstat_headers = []
        HpHpstat_headers = []

        for i in range(len(self.Lambdas)):
            MIstat_headers += ["MI_dir_"+h for h in self.HT_MI.getHeaders_1D(Lambda=self.Lambdas[i],ID=None)]
            HHstat_headers += ["HH_"+h for h in self.HT_H.getHeaders_1D2D(Lambda=self.Lambdas[i],IDs=None,ver=2)]
            HHjitstat_headers += ["HH_jit_"+h for h in self.HT_jit.getHeaders_1D(Lambda=self.Lambdas[i],ID=None)]
            HpHpstat_headers += ["HpHp_" + h for h in self.HT_Hp.getHeaders_1D2D(Lambda=self.Lambdas[i], IDs=None, ver=2)]
        return MIstat_headers + Hstat_headers + HHstat_headers + HHjitstat_headers + HpHpstat_headers



"""
    This file contains the Transformation class, which is uses to perform transformation on the feature vector before it is passed to the autoencoder.
    Different types of transformations are supported:
        1. Key based permuation
        2. Key based encryption
"""
import random
import string
import numpy as np

def generate_key(size):
    return ''.join(random.choice(string.ascii_uppercase + string.digits +  string.ascii_lowercase) for _ in range(size))

class Transformation:
    def __init__(self, tf_type, key):
        self.tf_type = tf_type
        self.key = key
    
    def transform(self, vector):
        if self.tf_type == "0":
            return vector
        elif self.tf_type == "1":
            return self.permute(vector)
        elif self.tf_type == "2":
            return self.encrypt(vector)

    def permute(self, vector):
        """
            Key Based Random Permutation (KBRP)
            Assumption - Length of Key < Length of Feature Vector
        """

        # 1. init
        n = len(vector)
        s = min(len(self.key), n)
        K = [ord(c) for c in self.key]
        P = [i for i in range(1, n+1)]
        A = [0 for i in range(n)]

        for i in range(s):
            A[i] = K[i]
        for i in range(0, s-1):
            P[i] = A[i] + A[i+1] # According to text not pseudocode
        P[s-1] = A[0]
        for i in range(s, n):
            j = s
            while (j>0):
                for k in range(i-s, i):
                    if (j>=n):
                        break
                    P[i] = P[i] + P[k]
                    j -= 1

    
        for i in range(n):
            P[i] = P[i] % n

        # 2. eliminate
        l = 0
        r = n-1
        while (l<r):
            for i in range(l+1, r+1):
                if (P[i] == P[l]):
                    P[i] = 0
            l += 1
            for i in range(r-1, l-1, -1):
                if (P[i] == P[r]):
                    P[i] = 0
            r -= 1


        # 3. fill
        missing = set([i+1 for i in range(len(P))]) - set(P)
        m = len(missing)
        
        l = 0
        r = n-1
        side = 1
        while (len(missing) != 0):
            if side:
                while (l < len(P) and P[l] != 0):
                    l += 1
                if (l == len(P)):
                    break
                P[l] = missing.pop()
                side = 0
            else:
                while (r > 0 and P[r] != 0):
                    r -= 1
                if (r == -1):
                    break
                P[r] = missing.pop()
                side = 1

        # 4. permute vector

        return np.array([vector[i-1] for i in P])

    def encrypt(self, vector):
        return vector


"""
    Test Code
"""


if __name__ == "__main__":
    t = Transformation("1", "hello")
    p = t.permute([i+1 for i in range(0, 30)])

    # check if permutation is valid
    assert len(p) == 30
    assert set(p) == set([i+1 for i in range(0, 30)])





# from .kitsune import KitModel, KeyedKitModel
from .awesome_ids import awesome_ids_model
# do we have to make any changes here to create the function inside awesome ids , I dont want to call kitsune and keyed kitsune here


__all__ = [
    "AwesomeIDS"
    ]


# Export the public API for the kitsune package.
from .kit_model import KitModel
from .kit_model_keyed import KeyedKitModel
from .metrics import evaluation_metrics

__all__ = [
    "KitModel",
    "KeyedKitModel",
    "evaluation_metrics"
    ]
